/* bitmap.c: print the cluster bitmap of a file: `1' for compressed
   clusters, `0' for uncompressed clusters. */

#include <linux/autoconf.h>
#ifndef CONFIG_EXT2_COMPRESS
/* A few things need to be defined for <linux/ext2_fs.h> to give
   us what we want. */
# define CONFIG_EXT2_COMPRESS
# define CONFIG_EXT2_DEFAULT_COMPR_METHOD_GZIP9 /* Change this if you want. */
# define CONFIG_EXT2_DEFAULT_CLUSTER_BITS_5     /* Change this if you want. */
# define CONFIG_EXT2_USE_GZIP /* prevents ext2_fs.h (e2compr-0.3.x) from giving #error. */
# define CONFIG_EXT2_HAVE_GZIP /* prevents ext2_fs_c.h (e2compr-0.4.x) from giving #error. */
#endif

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <linux/fs.h>  /* FIGETBSZ */
#include <linux/ext2_fs.h>
#include <errno.h>
#include <string.h>

static char *prog_name;

static char *version_string = "e2bitmap 1998-06-19";

static void
usage (int rc)
{
  fprintf (stderr,
	   "Usage: %s <files>\n"
	   "Shows the cluster bitmap for each regular file on the command line.\n",
	   prog_name);
  exit(rc);
}


int
main (int argc, char **argv)
{
  int rc = 0;
  prog_name = *argv++;

  if (argc == 1)
    {
      (void) fprintf (stderr, "%s: no arguments.\nTry `--help'.\n",
		      prog_name);
      exit (EXIT_FAILURE);
    }

  if (strcmp (*argv, "--help") == 0)
    {
      puts (version_string);
      usage(0);
    }

  if (strcmp (*argv, "--version") == 0)
    {
      puts (version_string);
      exit(0);
    }

  if (**argv == '-')
    {
      if (strcmp (*argv, "--") == 0)
	argv++;
      else
	{
	  (void) fprintf (stderr,
			  "%s: unrecognised option `%s'\nTry `--help'.\n",
			  prog_name, *argv);
	  exit(1);
	}
    }
	  
  for (; *argv != NULL; ++argv)
    {
      int ifd;
      unsigned long value, clu_nblocks, blk_nchars;
      int i, nclusters, ncompressed;
      struct stat st;

      if ((lstat (*argv, &st) == -1)
	  || (!S_ISREG (st.st_mode)))
	{
	  /* Ignore it. */
	  continue;
	}

      ifd = open(*argv, O_RDONLY);
      if ((ifd == -1)
	  || (ioctl(ifd, EXT2_IOC_GETFLAGS, &value) == -1))
	{
	  perror (prog_name);
	  if (ifd != -1)
	    close(ifd);
	  rc |= 1;
	  continue;
	}

      if ((value & EXT2_COMPRBLK_FL) == 0)
	{
	  printf ("%s: uncompressed\n", *argv);
	  close (ifd);
	  continue;
	}

      if (ioctl(ifd, EXT2_IOC_GETCLUSTERSIZE, &clu_nblocks) == -1)
	{
	  (void) fprintf (stderr, "%s: can't read cluster size on `%s': %s\n",
			  prog_name, *argv, strerror(errno));
	  rc |= 1;
	  (void) close (ifd);
	  continue;
	}

      if (ioctl (ifd, FIGETBSZ, &blk_nchars))
	{
	  (void) fprintf (stderr, "%s: can't read block size on `%s': %s\n",
			  prog_name, *argv, strerror(errno));
	  rc |= 1;
	  (void) close (ifd);
	  continue;
	}

      nclusters = (st.st_size
		   ? ((st.st_size - 1) / (clu_nblocks * blk_nchars)) + 1
		   : 0);
      ncompressed = 0;
      printf ("%s: ", *argv);
      for (i = 0; i < nclusters; i++)
	{
	  value = i;
	  ioctl(ifd, EXT2_IOC_GETCLUSTERBIT, &value);
	  if (value)
	    {
	      putchar ('1');
	      ncompressed++;
	    }
	  else
	    putchar ('0');
	}
      printf (" (%.2f%%)\n", 100.0 * ncompressed / nclusters);
      close (ifd);
    }

  return rc;
}
