/*
   Comments from section 12.8 "readn and writen Functions", Stevens, "Advanced
   Programming in the UNIX Environment".
   _____________________________________________________________________________

   Some devices, notably terminals, networks, and any SVR4 streams devices, have
   the following two properties.

      1.  A read operation may return less than asked for, even though we have
          not encountered the end of file.  This is not an error, and we should
          just continue reading from the device.

      2.  A write operation can also return less than we specified.  This may be
          caused by flow control constraints by downstream modules, for example.
          Again, it's not an error, and we should continue writing the remainder
          of the data.  (Normally this short return from a write only occurs
          with a nonblocking descriptor or if a signal is caught.)

   We'll never see this happen when reading or writing a disk file.

   ...

   We call writen anytime we're writing to one of the device types that we
   mentioned, but we call readn only when we know ahead of time that we will be
   receiving a certain number of bytes.  (Often we issue a read to one of the
   devices and take whatever is returned.)
   _____________________________________________________________________________

 */

#include <sys/types.h>
#include <unistd.h>
#include <assert.h>
#include <errno.h>

ssize_t
writen (int fd, const void *vptr, size_t n)
{
  size_t nleft;
  ssize_t nwritten;
  const char *ptr;

  ptr = vptr;  /* Can't do pointer arithmetic on void* . */
  nleft = n;
  while (nleft > 0)
    {
#if 0  /* This is what Stevens had. */
      if ((nwritten = write (fd, ptr, nleft)) <= 0)
	return (nwritten);	/* error */

      nleft -= nwritten;
      ptr += nwritten;
#else
      if ((nwritten = write (fd, ptr, nleft)) < 0)
	{
	  assert (nwritten == -1);
	  switch (errno)
	    {
	    case EINTR:
	    case EAGAIN:
	      continue;
	    default:
	      return nwritten;
	    }
	}
      else
	{
	  nleft -= nwritten;
	  ptr += nwritten;
	}
#endif /* 0 */
    }

  return n;
}


ssize_t
readn (int fd, void *vptr, size_t n)
{
  size_t nleft;
  ssize_t nread;
  char *ptr;

  ptr = vptr;
  nleft = n;
  while (nleft > 0)
    {
#if 0 /* Stevens */
      if ((nread = read (fd, ptr, nleft)) < 0)
	return (nread);		/* error, return < 0 */
      else if (nread == 0)
	break;

      nleft -= nread;
      ptr += nread;
#else
      if ((nread = read (fd, ptr, nleft)) < 0)
	{
	  assert (nread == -1);
	  switch (errno)
	    {
	    case EINTR:
	    case EAGAIN:
	      continue;
	    default:
	      return nread;
	    }
	}
      else if (nread == 0)
	break;
      else
	{
	  nleft -= nread;
	  ptr += nread;
	}
#endif
    }
  return (n - nleft);		/* return >= 0 */
}
